#!/usr/bin/python3

import os,sys
import time

import cv2
import numpy as np
import SimpleITK as sitk
import matplotlib.pyplot as plt
from ReadSagittalPD import ReadSagittalPDs

from PyQt5.QtWidgets import QGridLayout, \
        QDesktopWidget, QMainWindow, QApplication, qApp, \
        QDockWidget, QWidget, QAction, QTableWidget, QTableWidgetItem, \
        QMenu, QPushButton, \
        QVBoxLayout, QHBoxLayout, \
        QFileDialog, QMessageBox, QDialog
from PyQt5.QtCore import Qt, pyqtSlot, QSize

from volumeViewerWidget import volumeViewerWidget, volumeSliceViewerWidget
from dicomInfoWidget import dicomInfoWidget
from gradeDispWidget import gradeDispWidget

from mainLogic import mainLogic

class dicomSelectDialog(QDialog):
    def __init__(self, images, parent=None):
        super(dicomSelectDialog, self).__init__(parent)
        self.__images = images
        self.btnOK = QPushButton('OK')
        self.btnQuit = QPushButton('Cancel')
        self.table = QTableWidget(len(images), 3)
        hbox = QHBoxLayout()
        hbox.addStretch(1)
        hbox.addWidget(self.btnOK)
        hbox.addWidget(self.btnQuit)
        vbox = QVBoxLayout(self)
        vbox.addWidget(self.table, stretch=1)
        vbox.addLayout(hbox)
        self.setLayout(vbox)
        self.__makeTable()
        self.selectedIndex = None
        
        self.btnQuit.clicked.connect(self.reject)
        #self.btnOK.clicked.connect( \
        #        lambda x, self=self: self.done(self.table.currentRow()))
        self.btnOK.clicked.connect(self.btnOKClicked)

    def btnOKClicked(self, flag):
        self.selectedIndex = self.table.currentRow()
        self.accept()

    def __makeTable(self):
        labels = ['PatientID', 'SeriesDescription', 'Size']
        self.table.setHorizontalHeaderLabels(labels)
        for i, image in enumerate(self.__images):
            self.table.setItem(i, 0, \
                    QTableWidgetItem(str(image.GetMetaData('0010|0020'))))
            self.table.setItem(i, 1, \
                    QTableWidgetItem(str(image.GetMetaData('0008|103e'))))
            self.table.setItem(i, 2, \
                    QTableWidgetItem(str(image.GetSize())))
        self.table.resizeColumnsToContents()
        self.table.adjustSize()


class controlPannelWidget(QWidget):
    def __init__(self, parent=None):
        super(controlPannelWidget, self).__init__(parent)
        self.btnLoad = QPushButton('Load Study')
        #self.btnLoad.setToolTip('Load a dicom study from directory.')
        self.btnCla = QPushButton('Predict Grade')
        #self.btnCla.setToolTip('Predict grade of selected image')
        self.btnRef = QPushButton('Get Reference')
        #self.btnRef.setToolTip('What a healthy knee joint looks like?')
        self.btnSeg = QPushButton('Auto Segment')
        #self.btnSeg.setToolTip(\
        #        'Perform automated segmentation of Tibia/Femur/Cartilage.')
        vbox = QVBoxLayout(self)
        vbox.addWidget(self.btnLoad)
        vbox.addWidget(self.btnCla)
        vbox.addWidget(self.btnSeg)
        vbox.addWidget(self.btnRef)
        self.setLayout(vbox)


class mainWindow(QMainWindow):
    def __init__(self):
        super(mainWindow, self).__init__()
        self.__image = None # SimpleITK.Image
        self.__grade = None # Numpy array with shape [num_slice, num_class]
        self.statusBar().showMessage('Initializing UI...')
        self.initUI()
        self.statusBar().showMessage('Initializing Core...')
        self.main = mainLogic()
        self.statusBar().showMessage('Initializing Signal...')
        self.initSignals()
        self.statusBar().showMessage('Ready')
    
    def initUI(self):
        self.setWindowTitle('Cartilage Evaluation System')
        screenGeometry = QApplication.desktop().screenGeometry()
        aspectRatio = 4/3.0
        blockLen = min(screenGeometry.width()/aspectRatio, \
                screenGeometry.height())
        self.resize(blockLen*aspectRatio, blockLen)
        #qr = self.frameGeometry()
        #cp = QDesktopWidget().availableGeometry().center()
        #qr.moveCenter(cp)
        #self.move(qr.topLeft())
        self.initUI_Menubar()
        self.initUI_Toolbar()
        self.initUI_ContextMenu()
        self.initUI_Layout()
        self.show()

    def initUI_Menubar(self):
        menubar = self.menuBar()
        #fileMenu = menubar.addMenu('File')
        #loadStudyAct = QAction('Load Dicom Study', self)
        #fileMenu.addAction(loadStudyAct)
        #exitAct = QAction('Quit', self)
        #exitAct.triggered.connect(qApp.quit)
        #fileMenu.addAction(exitAct)

    def initUI_Toolbar(self):
        pass

    def initUI_ContextMenu(self):
        pass

    def initUI_Layout(self):
        self.volumeViewer = volumeViewerWidget(self)
        self.dicomInfo = dicomInfoWidget(self)
        self.controlPanel = controlPannelWidget(self)
        self.gradeDisp = gradeDispWidget(self)
        self.refViewer = volumeSliceViewerWidget(self)
        colormap = (plt.cm.bwr(np.array( \
                list([x for x in range(256)]), dtype=np.uint8 \
                ))*255).astype(np.uint8).reshape(256,1,4) #RGBA
        colormap[:,0,3] = list(range(255,0,-2))+list(range(1,256,2))
        self.resViewer = volumeSliceViewerWidget( \
                self, labelColormap=colormap)
        # dock and central widget
        # volumeViewer
        self.setCentralWidget(self.volumeViewer)
        # dicomInfo
        self.dockDicomInfo = QDockWidget("DICOM Info", self)
        self.dockDicomInfo.setObjectName("dockDicomInfo")
        self.addDockWidget(Qt.RightDockWidgetArea, self.dockDicomInfo)
        self.dockDicomInfo.setWidget(self.dicomInfo)
        # gradeDisp
        self.dockGradeDisp = QDockWidget("Grade (Prediction)", self)
        self.dockGradeDisp.setObjectName("dockGradeDisp")
        self.addDockWidget(Qt.RightDockWidgetArea, self.dockGradeDisp)
        self.dockGradeDisp.setWidget(self.gradeDisp)
        self.tabifyDockWidget(self.dockDicomInfo, self.dockGradeDisp)
        # controlPanel
        self.dockControlPanel = QDockWidget("Control Panel", self)
        self.dockControlPanel.setObjectName("dockControlPanel")
        self.addDockWidget(Qt.RightDockWidgetArea, self.dockControlPanel)
        self.dockControlPanel.setWidget(self.controlPanel)
        self.dockControlPanel.showMinimized()
        self.resizeDocks([self.dockControlPanel, self.dockGradeDisp], \
                [1,10], Qt.Vertical)
        # reference image and residual image
        self.dockRefViewer = QDockWidget("Reference", self)
        self.dockRefViewer.setObjectName("dockRefViewer")
        self.addDockWidget(Qt.LeftDockWidgetArea, self.dockRefViewer)
        self.dockRefViewer.setWidget(self.refViewer)
        self.dockRefViewer.setFloating(True)
        self.dockRefViewer.setVisible(False)
        self.dockResViewer = QDockWidget("Residual", self)
        self.dockResViewer.setObjectName("dockResViewer")
        self.addDockWidget(Qt.LeftDockWidgetArea, self.dockResViewer)
        self.dockResViewer.setWidget(self.resViewer)
        self.dockResViewer.setFloating(True)
        self.dockResViewer.setVisible(False)


    def initSignals(self):
        self.controlPanel.btnLoad.clicked.connect(self.actLoadStudy)
        self.controlPanel.btnCla.clicked.connect(self.actGetGrade)
        self.controlPanel.btnRef.clicked.connect(self.actGetReference)
        self.controlPanel.btnSeg.clicked.connect(self.actGetSeg)
        self.volumeViewer.sliderIndex.valueChanged.connect( \
                lambda x: self.refViewer.setIndex(x-1))
        self.volumeViewer.sliderIndex.valueChanged.connect( \
                lambda x: self.resViewer.setIndex(x-1))
        self.gradeDisp.cellPressed.connect( \
                lambda row, col: self.volumeViewer.sliderIndex.setValue(row+1))

    @pyqtSlot()
    def actLoadStudy(self, directory=None):
        start = time.time()
        self.statusBar().showMessage('Loading Study...')
        if directory is None:
            dialog = QFileDialog(self) 
            dialog.setFileMode(QFileDialog.DirectoryOnly)
            dialog.setViewMode(QFileDialog.List)
            dialog.setOption(QFileDialog.ShowDirsOnly, True)
            if dialog.exec_():
                directory = str(dialog.selectedFiles()[0])
            else:
                self.statusBar().showMessage( \
                        'Ready ({:.2f}s)'.format(time.time() - start))
                return
        try:
            images = ReadSagittalPDs(directory)
        except Exception as err:
            msgBox = QMessageBox(self)
            msgBox.setText(str(type(err)) + str(err))
            msgBox.exec()
            self.statusBar().showMessage( \
                    'Ready ({:.2f}s)'.format(time.time() - start))
            return
        selectDialog = dicomSelectDialog(images)
        if selectDialog.exec_():
            image = images[selectDialog.selectedIndex]
        else:
            self.statusBar().showMessage( \
                    'Ready ({:.2f}s)'.format(time.time() - start))
            return
        self.main.setImage(image)
        self.volumeViewer.setImage(image)
        self.dicomInfo.setImage(image)
        self.dockDicomInfo.setVisible(True)
        self.dockRefViewer.setVisible(False)
        self.dockResViewer.setVisible(False)
        self.gradeDisp.setGrade(None)
        self.statusBar().showMessage( \
                'Ready ({:.2f}s)'.format(time.time() - start))
        return
    
    @pyqtSlot()
    def actGetGrade(self):
        start = time.time()
        self.statusBar().showMessage('Predicting Grade...')
        try:
            grade = self.main.getGrade()
        except Exception as err:
            msgBox = QMessageBox(self)
            msgBox.setText(str(type(err)) + str(err))
            msgBox.exec()
            self.statusBar().showMessage( \
                    'Ready ({:.2f}s)'.format(time.time() - start))
            return
        self.gradeDisp.setGrade(grade)
        self.dockGradeDisp.setVisible(True)
        self.statusBar().showMessage( \
                'Ready ({:.2f}s)'.format(time.time() - start))
        msgBox = QMessageBox(self)
        msgBox.setText('得出诊断意见：'+
                '我们从四千多个膝关节测共振图层（健康人和患者）'+\
                '与医生的诊断结果中，利用深度学习的方法，'+\
                '提取图像的特征及其与诊断结果的关系。'+\
                '利用训练出的模型，'+\
                '计算机可以自动从图像中预测诊断意见。'+'\n'+\
                '提示水肿区域：'+\
                '我们从三百多层带有水肿区域的磁共振图像'+\
                '与医生手工勾勒出的水肿区域中，'+\
                '利用深度学习的方法提取特征。利用训练出的模型，'+\
                '计算机可以自动从图像中预测图像是否有水肿，'+\
                '以及水肿区域的位置。')
        msgBox.exec()
        return

    @pyqtSlot()
    def actGetReference(self):
        start = time.time()
        self.statusBar().showMessage('Predicting Reference...')
        try:
            swollen = self.main.getSwollen()
            self.main.getInpainted()
            self.main.getSuperResolution()
            seg = self.main.getSeg()
            ref = self.main.getReference()
            img = self.main.getImage()
            maxVal = sitk.GetArrayFromImage(img).max()
            scale = 3
            res = sitk.Cast(ref, sitk.sitkFloat32) - \
                    sitk.Cast(img, sitk.sitkFloat32)
            res = res * \
                    sitk.Cast((seg==2)+(seg==4)+(seg==6), sitk.sitkFloat32) * \
                    (scale*255.0) / maxVal
            res = sitk.Cast(res + 127.5, sitk.sitkUInt8)
        except Exception as err:
            msgBox = QMessageBox(self)
            msgBox.setText(str(type(err)) + str(err))
            msgBox.exec()
            self.statusBar().showMessage( \
                    'Ready ({:.2f}s)'.format(time.time() - start))
            return
        self.refViewer.setImage(ref)
        self.refViewer.setIndex(self.volumeViewer.sliderIndex.value()-1)
        self.resViewer.setImage(img)
        self.resViewer.setLabel(res)
        self.resViewer.setIndex(self.volumeViewer.sliderIndex.value()-1)
        if not self.dockRefViewer.isVisible():
            self.dockRefViewer.setVisible(True)
            if self.dockRefViewer.isFloating():
                self.dockRefViewer.resize( \
                        self.refViewer.viewportSizeHint()*1.1)
                #self.refViewer.resize(self.refViewer.viewportSizeHint())
        if not self.dockResViewer.isVisible():
            self.dockResViewer.setVisible(True)
            if self.dockResViewer.isFloating():
                self.dockResViewer.resize( \
                        self.resViewer.viewportSizeHint()*1.1)
        self.volumeViewer.setLabel(swollen)
        self.statusBar().showMessage( \
                'Ready ({:.2f}s)'.format(time.time() - start))
        msgBox = QMessageBox(self)
        msgBox.setText('生成参考图像：'+\
                '我们从400个健康人的图像中训练出了一个模型'+\
                '（针对正常人磁共振图像提取出的先验知识），'+\
                '将这个模型应用到此图像中，可以得到正常情况下的参考图像')
        msgBox.exec()

    @pyqtSlot()
    def actGetSeg(self):
        start = time.time()
        self.statusBar().showMessage('Predicting Segmentation...')
        try:
            seg = self.main.getSeg()
        except Exception as err:
            msgBox = QMessageBox(self)
            msgBox.setText(str(type(err)) + str(err))
            msgBox.exec()
            self.statusBar().showMessage( \
                    'Ready ({:.2f}s)'.format(time.time() - start))
            return
        self.volumeViewer.setLabel(seg)
        self.statusBar().showMessage( \
                'Ready ({:.2f}s)'.format(time.time() - start))
        msgBox = QMessageBox(self)
        msgBox.setText('生成分割结果：'+\
                '我们从100多例患者膝关节磁共振的手工标记结果中，'+\
                '利用深度学习的方法提取特征。利用训练出的模型，'+\
                '计算机可以自动从图像中得到膝关节结构的分割结果。')
        msgBox.exec()

if __name__ == '__main__':
    dicomPath = '/mnt/repo/privateData/cartilage_origin/FromXuhua/PD/A102752442'
    if len(sys.argv) > 1:
        dicomPath = sys.argv[1]
    app = QApplication(sys.argv)
    mainWindow = mainWindow()
    mainWindow.actLoadStudy(dicomPath)
    sys.exit(app.exec_())

